"""Combobox component for Mantine UI."""

from typing import Any, Literal

import reflex as rx
from reflex.components.component import Component
from reflex.components.props import PropsBase

from reflex_enterprise.components.mantine.base import BaseMantineInput, MantineCoreBase


class ComboboxBase(MantineCoreBase):
    """Base class for Combobox components."""


class MiddlewaresProps(PropsBase):
    """Props for dropdown positioning middlewares."""

    shift: rx.Var[bool]
    flip: rx.Var[bool]
    inline: rx.Var[bool]


class Combobox(
    BaseMantineInput,
    ComboboxBase,
):
    """Displays a customizable dropdown list."""

    tag = "Combobox"

    # Arrow offset in px, 5 by default
    arrow_offset: rx.Var[int]

    # Arrow position
    arrow_position: rx.Var[Literal["center", "side"]]

    # Arrow border-radius in px, 0 by default
    arrow_radius: rx.Var[int]

    # Arrow size in px, 7 by default
    arrow_size: rx.Var[int]

    # If set, popover dropdown will not be rendered
    disabled: rx.Var[bool]

    # Controls padding of the dropdown, 4 by default
    dropdown_padding: rx.Var[str | int]

    # Changes floating ui position strategy, 'absolute' by default
    floating_strategy: rx.Var[Literal["absolute", "fixed"]]

    # If set, the dropdown is hidden when the element is hidden with styles or not visible on the screen, true by default
    hide_detached: rx.Var[bool]

    # If set dropdown will not be unmounted from the DOM when it is hidden, display: none styles will be added instead
    keep_mounted: rx.Var[bool]

    # Floating ui middlewares to configure position handling, { flip: true, shift: true, inline: false } by default
    middlewares: rx.Var[MiddlewaresProps | dict[str, Any]]

    # Offset of the dropdown element, 8 by default
    offset: rx.Var[int]

    # Props passed down to Overlay component
    overlay_props: rx.Var[dict[str, Any]]

    # Props to pass down to the Portal when withinPortal is true
    portal_props: rx.Var[dict[str, Any]]

    # Dropdown position relative to the target element, 'bottom' by default
    position: rx.Var[Literal["top", "bottom", "left", "right"]]

    # useEffect dependencies to force update dropdown position, [] by default
    position_dependencies: rx.Var[list[Any]]

    # Key of theme.radius or any valid CSS value to set border-radius, theme.defaultRadius by default
    radius: rx.Var[str | int]

    # Determines whether selection should be reset when option is hovered, false by default
    reset_selection_on_option_hover: rx.Var[bool]

    # Determines whether focus should be automatically returned to control when dropdown closes, false by default
    return_focus: rx.Var[bool]

    # Key of theme.shadows or any other valid CSS box-shadow value
    shadow: rx.Var[str]

    # Combobox store, can be used to control combobox state
    store: rx.Var[Any] = rx.Var.create("combobox")

    # Props passed down to the Transition component that used to animate dropdown presence, use to configure duration and animation type, { duration: 150, transition: 'fade' } by default
    transition_props: rx.Var[dict[str, Any]]

    # Dropdown width, or 'target' to make dropdown width the same as target element, 'max-content' by default
    width: rx.Var[str | int | Literal["target"]]

    # Determines whether component should have an arrow, false by default
    with_arrow: rx.Var[bool]

    # Determines whether the overlay should be displayed when the dropdown is opened, false by default
    with_overlay: rx.Var[bool]

    # Determines whether dropdown should be rendered within the Portal, true by default
    with_portal: rx.Var[bool]

    # Dropdown z-index, 300 by default
    z_index: rx.Var[int]

    # Called when dropdown closes
    on_close: rx.EventHandler[rx.event.no_args_event_spec]

    # Called when the popover is dismissed by clicking outside or by pressing escape
    on_dismiss: rx.EventHandler[rx.event.no_args_event_spec]

    # Called when enter transition ends
    on_enter_transition_end: rx.EventHandler[rx.event.no_args_event_spec]

    # Called when exit transition ends
    on_exit_transition_end: rx.EventHandler[rx.event.no_args_event_spec]

    # Called when dropdown opens
    on_open: rx.EventHandler[rx.event.no_args_event_spec]

    # Called when item is selected with Enter key or by clicking it
    on_option_submit: rx.EventHandler[
        rx.event.passthrough_event_spec(tuple[str, dict[str, Any]])
    ]

    # Called when dropdown position changes
    on_position_change: rx.EventHandler[rx.event.passthrough_event_spec(float)]

    def add_hooks(self):
        """Add hooks to the component."""
        combobox_hook = rx.Var(
            "const combobox = useCombobox()",
            _var_data=rx.vars.VarData(
                imports={"@mantine/core": "useCombobox"},
                position=rx.constants.Hooks.HookPosition.PRE_TRIGGER,
            ),
        )
        return [combobox_hook]


class ComboboxOptions(ComboboxBase):
    """Combobox options element wrapper."""

    tag = "Combobox.Options"


class ComboboxOption(ComboboxBase):
    """Combobox option."""

    tag = "Combobox.Option"

    # If set, highlights the option as active
    active: rx.Var[bool]

    # If set, option cannot be selected
    disabled: rx.Var[bool]

    # Determines whether item is selected, useful for virtualized comboboxes
    selected: rx.Var[bool]

    # Option value
    value: rx.Var[str]


class ComboboxTarget(ComboboxBase):
    """Target element wrapper, wraps the element that opens dropdown."""

    tag = "Combobox.Target"

    _rename_props = {"target": "children"}

    # Input autocomplete attribute
    autocomplete: rx.Var[str]

    # Target element
    target: rx.Var[Any]

    # Key of the prop that should be used to access element ref
    ref_prop: rx.Var[str]

    # Determines which events should be handled by the target element. button target type handles Space and Enter keys to toggle dropdown opened state. input by default.
    target_type: rx.Var[Literal["button", "input"]]

    # Determines whether the target should have aria- attributes, true by default
    with_aria_attributes: rx.Var[bool]

    # Determines whether the target should have aria-expanded attribute, false by default
    with_expanded_attribute: rx.Var[bool]

    # Determines whether component should respond to keyboard events, true by default
    with_keyboard_navigation: rx.Var[bool]


class ComboboxDropdownTarget(ComboboxBase):
    """Dropdown target element wrapper, wraps the element that opens dropdown."""

    tag = "Combobox.DropdownTarget"

    _rename_props = {"target": "children"}

    # Dropdown target element
    target: rx.Var[Any]

    # Key of the prop that should be used to access element ref
    ref_prop: rx.Var[str]


class ComboboxEventsTarget(ComboboxTarget):
    """Events target element wrapper, wraps the element that opens dropdown."""

    tag = "Combobox.EventsTarget"


class ComboboxDropdown(ComboboxBase):
    """Dropdown element wrapper."""

    tag = "Combobox.Dropdown"

    # If set, dropdown will be hidden, useful for custom transitions
    hidden: rx.Var[bool]


class ComboboxGroup(ComboboxBase):
    """Combobox group element wrapper."""

    tag = "Combobox.Group"

    # Group label
    label: rx.Var[Component | str]


class ComboboxNamespace(rx.ComponentNamespace):
    """Namespace for Combobox components."""

    __call__ = staticmethod(Combobox.create)
    options = staticmethod(ComboboxOptions.create)
    option = staticmethod(ComboboxOption.create)
    target = staticmethod(ComboboxTarget.create)
    dropdown_target = staticmethod(ComboboxDropdownTarget.create)
    events_target = staticmethod(ComboboxEventsTarget.create)
    dropdown = staticmethod(ComboboxDropdown.create)
    group = staticmethod(ComboboxGroup.create)


combobox = ComboboxNamespace()



"""Flow component for managing the flow of data between nodes."""

from typing import TYPE_CHECKING, Any, Literal, Mapping, Sequence, TypedDict

import reflex
from reflex.components.el.elements.base import AriaRole
from typing_extensions import NotRequired


class XYPosition(TypedDict):
    """All positions are stored in an object with x and y coordinates."""

    x: float
    y: float


class ProOptions(TypedDict):
    """By default, we render a small attribution in the corner of your flows that links back to the project.

    Anyone is free to remove this attribution whether they’re a Pro subscriber or not but we ask that you take a quick look at our removing attribution guide before doing so.
    """

    account: NotRequired[str]
    hideAttribution: bool


PanelPosition = Literal[
    "top-left",
    "top-center",
    "top-right",
    "bottom-left",
    "bottom-center",
    "bottom-right",
]

AriaLabelConfig = TypedDict(
    "AriaLabelConfig",
    {
        "node.a11yDescription.default": NotRequired[str],
        "node.a11yDescription.keyboardDisabled": NotRequired[str],
        "node.a11yDescription.ariaLiveMessage": NotRequired[Any],
        "edge.a11yDescription.default": NotRequired[str],
        "controls.ariaLabel": NotRequired[str],
        "controls.zoomIn.ariaLabel": NotRequired[str],
        "controls.zoomOut.ariaLabel": NotRequired[str],
        "controls.fitView.ariaLabel": NotRequired[str],
        "controls.interactive.ariaLabel": NotRequired[str],
        "minimap.ariaLabel": NotRequired[str],
        "handle.ariaLabel": NotRequired[str],
    },
)

SnapGrid = tuple[float, float]

Position = Literal["top", "right", "bottom", "left"]

CoordinateExtent = tuple[tuple[float, float], tuple[float, float]]

NodeOrigin = tuple[float, float]

HandleType = Literal["source", "target"]

ColorMode = Literal["light", "dark", "system"]


class Viewport(TypedDict):
    """Internally, React Flow maintains a coordinate system that is independent of the rest of the page.

    The Viewport type tells you where in that system your flow is currently being display at and how zoomed in or out it is.
    """

    x: float
    y: float
    zoom: float


class Rect(TypedDict):
    """The Rect type defines a rectangle in a two-dimensional space with dimensions and a position."""

    x: float
    y: float
    width: float
    height: float


class NodeHandle(TypedDict):
    """The NodeHandle type is used to define a handle for a node if server-side rendering is used.

    On the server, React Flow can’t measure DOM nodes, so it’s necessary to define the handle position dimensions.
    """

    width: NotRequired[float]
    height: NotRequired[float]
    id: NotRequired[str | None]
    x: float
    y: float
    pos: Position
    type: HandleType


class Measured(TypedDict):
    """Represents the measured dimensions of a node."""

    width: NotRequired[float]
    height: NotRequired[float]


NodeType = Literal["default", "input", "output", "group"]


class Id(TypedDict):
    """Represents a unique identifier for a node."""

    id: str


class Padding(TypedDict):
    """Represents the padding values for a node."""

    top: NotRequired[float]
    right: NotRequired[float]
    bottom: NotRequired[float]
    left: NotRequired[float]
    x: NotRequired[float]
    y: NotRequired[float]


class Node(TypedDict):
    """The Node type represents everything React Flow needs to know about a given node.

    Many of these properties can be manipulated both by React Flow or by you, but some such as width and height should be considered read-only.
    """

    id: str
    """Unique id of a node."""

    position: XYPosition
    """Position of a node on the pane."""

    data: NotRequired[dict[str, Any]]
    """Arbitrary data passed to a node."""

    sourcePosition: NotRequired[Position]
    """Only relevant for default, source, target nodeType. Controls source position."""

    targetPosition: NotRequired[Position]
    """Only relevant for default, source, target nodeType. Controls target position."""

    hidden: NotRequired[bool]
    """Whether or not the node should be visible on the canvas."""

    selected: NotRequired[bool]

    dragging: NotRequired[bool]
    """Whether or not the node is currently being dragged."""

    draggable: NotRequired[bool]
    """Whether or not the node is able to be dragged."""

    selectable: NotRequired[bool]

    connectable: NotRequired[bool]

    deletable: NotRequired[bool]

    dragHandle: NotRequired[str]
    """A class name that can be applied to elements inside the node that allows those elements to act as drag handles, letting the user drag the node by clicking and dragging on those elements."""

    width: NotRequired[float]

    height: NotRequired[float]

    initialWidth: NotRequired[float]

    initialHeight: NotRequired[float]

    parentId: NotRequired[str]
    """Parent node id, used for creating sub-flows."""

    zIndex: NotRequired[int]
    """The z-index of the node."""

    extent: NotRequired[CoordinateExtent | Literal["parent"]]
    """The boundary a node can be moved in."""

    expandParent: NotRequired[bool]
    """When true, the parent node will automatically expand if this node is dragged to the edge of the parent node’s bounds."""

    ariaLabel: NotRequired[str]
    """Aria label for the node."""

    origin: NotRequired[NodeOrigin]
    """Origin of the node relative to its position."""

    handles: NotRequired[Sequence[NodeHandle]]

    measured: NotRequired[Measured]

    type: NotRequired[str | NodeType]
    """Type of node defined in nodeTypes."""

    style: NotRequired[Any]

    className: NotRequired[str]

    resizing: NotRequired[bool]

    focusable: NotRequired[bool]

    ariaRole: NotRequired[AriaRole]
    """The ARIA role attribute for the node element, used for accessibility."""

    domAttributes: NotRequired[dict[str, Any]]
    """General escape hatch for adding custom attributes to the node’s DOM element."""


class NodeHandleBounds(TypedDict):
    """Represents the bounds of a node's handles."""

    source: list[NodeHandle]
    target: list[NodeHandle]


class NodeBounds(XYPosition):
    """Represents the bounds of a node."""

    width: float
    height: float


class NodeInternals(TypedDict):
    """Represents the internal state of a node."""

    positionAbsolute: XYPosition
    z: float
    userNode: Node
    handleBounds: NotRequired[NodeHandleBounds]
    bounds: NotRequired[NodeBounds]


class InternalNode(Node):
    """The InternalNode type is identical to the base Node type but is extended with some additional properties used internally by React Flow. Some functions and callbacks that return nodes may return an InternalNode."""

    internals: NodeInternals


class FitViewOptions(TypedDict):
    """When calling fitView these options can be used to customize the behavior.

    For example, the duration option can be used to transform the viewport smoothly over a given amount of time.
    """

    padding: NotRequired[str | float | Padding]
    includeHiddenNodes: NotRequired[bool]
    minZoom: NotRequired[float]
    maxZoom: NotRequired[float]
    duration: NotRequired[float]
    ease: NotRequired[Any]
    interpolate: NotRequired[Literal["smooth", "linear"]]
    nodes: NotRequired[Sequence[Node | Id]]


EdgeType = Literal["default", "straight", "step", "smoothstep", "simplebezier"]


class EdgeMarker(TypedDict):
    """Edges can optionally have markers at the start and end of an edge. The EdgeMarker type is used to configure those markers!

    Check the docs for MarkerType for details on what types of edge marker are available.
    """

    type: Literal["arrow", "arrowclosed"]
    color: NotRequired[str | None]
    width: NotRequired[float]
    height: NotRequired[float]
    markerUnits: NotRequired[str]
    orient: NotRequired[str]
    strokeWidth: NotRequired[float]


class Edge(TypedDict):
    """Where a Connection is the minimal description of an edge between two nodes, an Edge is the complete description with everything React Flow needs to know in order to render it."""

    id: str
    """Unique id of an edge."""

    type: NotRequired[EdgeType | str]
    """Type of edge defined in edgeTypes."""

    source: str
    """Id of source node."""

    target: str
    """Id of target node."""

    sourceHandle: NotRequired[str | None]
    """Id of source handle, only needed if there are multiple handles per node."""

    targetHandle: NotRequired[str | None]
    """Id of target handle, only needed if there are multiple handles per node."""

    animated: NotRequired[bool]

    hidden: NotRequired[bool]

    deletable: NotRequired[bool]

    selectable: NotRequired[bool]

    data: NotRequired[dict[str, Any]]
    """Arbitrary data passed to an edge."""

    selected: NotRequired[bool]

    markerStart: NotRequired[str | EdgeMarker]
    """Set the marker on the beginning of an edge."""

    markerEnd: NotRequired[str | EdgeMarker]
    """Set the marker on the end of an edge."""

    zIndex: NotRequired[float]

    ariaLabel: NotRequired[str]

    interactionWidth: NotRequired[float]
    """ReactFlow renders an invisible path around each edge to make them easier to click or tap on. This property sets the width of that invisible path."""

    label: NotRequired[Any]
    """The label or custom element to render along the edge. This is commonly a text label or some custom controls."""

    labelStyle: NotRequired[Any]
    """Custom styles to apply to the label."""

    labelShowBg: NotRequired[bool]

    labelBgStyle: NotRequired[Any]

    labelBgPadding: NotRequired[tuple[float, float]]

    labelBgBorderRadius: NotRequired[float]

    style: NotRequired[Any]

    className: NotRequired[str]

    reconnectable: NotRequired[bool | HandleType]
    """Determines whether the edge can be updated by dragging the source or target to a new node. This property will override the default set by the edgesReconnectable prop on the <ReactFlow /> component."""

    focusable: NotRequired[bool]

    ariaRole: NotRequired[AriaRole]
    """The ARIA role attribute for the edge, used for accessibility."""

    domAttributes: NotRequired[dict[str, Any]]
    """General escape hatch for adding custom attributes to the edge’s DOM element."""


class SmoothStepEdgePathOptions(TypedDict):
    """Options for customizing the path of a smooth step edge."""

    offset: NotRequired[float]
    borderRadius: NotRequired[float]


class SmoothStepEdge(Edge):
    """A smooth step edge is a type of edge that has a smooth transition between steps."""

    type: Literal["smoothstep"]  # pyright: ignore[reportIncompatibleVariableOverride]
    pathOptions: NotRequired[SmoothStepEdgePathOptions]


class BezierEdgePathOptions(TypedDict):
    """Options for customizing the path of a Bezier edge."""

    curvature: NotRequired[float]


class BezierEdge(Edge):
    """A Bezier edge is a type of edge that has a Bezier curve between two points."""

    type: Literal["bezier"]  # pyright: ignore[reportIncompatibleVariableOverride]

    pathOptions: NotRequired[BezierEdgePathOptions]


class DefaultEdgeOptions(TypedDict):
    """Many properties on an Edge are optional.

    When a new edge is created, the properties that are not provided will be filled in with the default values passed to the defaultEdgeOptions prop of the <ReactFlow /> component.
    """

    type: NotRequired[str | None]
    """Type of edge defined in edgeTypes."""

    animated: NotRequired[bool]

    hidden: NotRequired[bool]

    deletable: NotRequired[bool]

    selectable: NotRequired[bool]

    data: NotRequired[dict[str, Any]]
    """Arbitrary data passed to an edge."""

    markerStart: NotRequired[str | EdgeMarker]
    """Set the marker on the beginning of an edge."""

    markerEnd: NotRequired[str | EdgeMarker]
    """Set the marker on the end of an edge."""

    zIndex: NotRequired[float]

    ariaLabel: NotRequired[str]

    interactionWidth: NotRequired[float]
    """ReactFlow renders an invisible path around each edge to make them easier to click or tap on. This property sets the width of that invisible path."""

    label: NotRequired[Any]
    """The label or custom element to render along the edge. This is commonly a text label or some custom controls."""

    labelStyle: NotRequired[Any]
    """Custom styles to apply to the label."""

    labelShowBg: NotRequired[bool]

    labelBgStyle: NotRequired[Any]

    labelBgPadding: NotRequired[tuple[float, float]]

    labelBgBorderRadius: NotRequired[float]

    style: NotRequired[Any]

    className: NotRequired[str]

    reconnectable: NotRequired[bool | HandleType]
    """Determines whether the edge can be updated by dragging the source or target to a new node. This property will override the default set by the edgesReconnectable prop on the <ReactFlow /> component."""

    focusable: NotRequired[bool]

    ariaRole: NotRequired[AriaRole]
    """The ARIA role attribute for the edge, used for accessibility."""

    domAttributes: NotRequired[dict[str, Any]]
    """General escape hatch for adding custom attributes to the edge’s DOM element."""


class OnDeleteParams(TypedDict):
    """Parameters for the onDelete event."""

    nodes: list[Node]
    edges: list[Edge]


class OnDelete(TypedDict):
    """Triggered when nodes or edges are deleted."""

    params: OnDeleteParams


class Dimensions(TypedDict):
    """Represents the dimensions of a node."""

    width: float
    height: float


class NodeDimensionChange(TypedDict):
    """Triggered when a node's dimensions change."""

    id: str
    type: Literal["dimensions"]
    dimensions: NotRequired[Dimensions]
    resizing: NotRequired[bool]
    setAttributes: NotRequired[bool | Literal["width", "height"]]


class NodePositionChange(TypedDict):
    """Triggered when a node's position changes."""

    id: str
    type: Literal["position"]
    position: NotRequired[XYPosition]
    positionAbsolute: NotRequired[XYPosition]
    dragging: NotRequired[bool]


class NodeSelectionChange(TypedDict):
    """Triggered when a node's selection state changes."""

    id: str
    type: Literal["select"]
    selected: bool


class NodeRemoveChange(TypedDict):
    """Triggered when a node is removed."""

    id: str
    type: Literal["remove"]


class NodeAddChange(TypedDict):
    """Triggered when a node is added."""

    item: Node
    type: Literal["add"]
    index: NotRequired[int]


class NodeReplaceChange(TypedDict):
    """Triggered when a node is replaced."""

    id: str
    type: Literal["replace"]
    item: Node


NodeChange = (
    NodeDimensionChange
    | NodePositionChange
    | NodeSelectionChange
    | NodeRemoveChange
    | NodeAddChange
    | NodeReplaceChange
)


class Connection(TypedDict):
    """The Connection type is the basic minimal description of an Edge between two nodes.

    The addEdge util can be used to upgrade a Connection to an Edge.
    """

    source: str
    """The id of the node this connection originates from."""

    target: str
    """The id of the node this connection terminates at."""

    sourceHandle: str | None
    """When not null, the id of the handle on the source node that this connection originates from."""

    targetHandle: str | None
    """When not null, the id of the handle on the target node that this connection terminates at."""


class NodeConnection(Connection):
    """The NodeConnection type is an extension of a basic Connection that includes the edgeId."""

    edgeId: str


NoConnection = TypedDict(
    "NoConnection",
    {
        "inProgress": Literal[False],
        "isValid": None,
        "from": None,
        "fromHandle": None,
        "fromPosition": None,
        "fromNode": None,
        "to": None,
        "toHandle": None,
        "toPosition": None,
        "toNode": None,
    },
)

ConnectionInProgress = TypedDict(
    "ConnectionInProgress",
    {
        "inProgress": Literal[True],
        "isValid": bool | None,
        "from": XYPosition,
        "fromHandle": NodeHandle,
        "fromPosition": Position,
        "fromNode": Node,
        "to": XYPosition,
        "toHandle": NodeHandle | None,
        "toPosition": Position,
        "toNode": Node | None,
    },
)

ConnectionState = TypedDict(
    "ConnectionState",
    {
        "inProgress": bool,
        "isValid": bool | None,
        "from": XYPosition | None,
        "fromHandle": NodeHandle | None,
        "fromPosition": Position | None,
        "fromNode": Node | None,
        "to": XYPosition | None,
        "toHandle": NodeHandle | None,
        "toPosition": Position | None,
        "toNode": Node | None,
    },
)


class EdgeAddChange(TypedDict):
    """Triggered when an edge is added."""

    item: Edge
    type: Literal["add"]
    index: NotRequired[int]


class EdgeRemoveChange(TypedDict):
    """Triggered when an edge is removed."""

    id: str
    type: Literal["remove"]


class EdgeReplaceChange(TypedDict):
    """Triggered when an edge is replaced."""

    id: str
    item: Edge
    type: Literal["replace"]


class EdgeSelectionChange(TypedDict):
    """Triggered when an edge's selection state changes."""

    id: str
    type: Literal["select"]
    selected: bool


EdgeChange = EdgeAddChange | EdgeRemoveChange | EdgeReplaceChange | EdgeSelectionChange


class OnConnectStartParams(TypedDict):
    """Parameters for the onConnectStart event."""

    nodeId: str | None
    handleId: str | None
    handleType: HandleType | None


class ViewportVar(reflex.vars.ObjectVar[Viewport]):
    """Represents the viewport of the flow."""

    if TYPE_CHECKING:
        x: reflex.vars.NumberVar[float]
        y: reflex.vars.NumberVar[float]
        zoom: reflex.vars.NumberVar[float]


class RectVar(reflex.vars.ObjectVar[Rect]):
    """Represents a rectangle in the flow."""

    if TYPE_CHECKING:
        x: reflex.vars.NumberVar[float]
        y: reflex.vars.NumberVar[float]
        width: reflex.vars.NumberVar[float]
        height: reflex.vars.NumberVar[float]


class XYPositionVar(reflex.vars.ObjectVar[XYPosition]):
    """Represents the position of a node in the flow."""

    if TYPE_CHECKING:
        x: reflex.vars.NumberVar[float]
        y: reflex.vars.NumberVar[float]


class NodeHandleVar(reflex.vars.ObjectVar[NodeHandle]):
    """Represents a handle on a node in the flow."""

    if TYPE_CHECKING:
        width: reflex.vars.NumberVar[float]
        height: reflex.vars.NumberVar[float]
        id: reflex.vars.StringVar | None
        x: reflex.vars.NumberVar[float]
        y: reflex.vars.NumberVar[float]
        pos: reflex.vars.StringVar[Position]
        type: reflex.vars.StringVar[HandleType]


class MeasuredVar(reflex.vars.ObjectVar[Measured]):
    """Represents the measured properties of a node in the flow."""

    if TYPE_CHECKING:
        width: reflex.vars.NumberVar[float]
        height: reflex.vars.NumberVar[float]


class NodeVar(reflex.vars.ObjectVar[Node]):
    """Represents a node in the flow."""

    if TYPE_CHECKING:
        id: reflex.vars.StringVar
        position: XYPositionVar
        data: reflex.vars.ObjectVar[Mapping[str, Any]]
        sourcePosition: reflex.vars.StringVar[Position]  # noqa: N815
        targetPosition: reflex.vars.StringVar[Position]  # noqa: N815
        hidden: reflex.vars.BooleanVar
        selected: reflex.vars.BooleanVar
        dragging: reflex.vars.BooleanVar
        draggable: reflex.vars.BooleanVar
        selectable: reflex.vars.BooleanVar
        connectable: reflex.vars.BooleanVar
        deletable: reflex.vars.BooleanVar
        dragHandle: reflex.vars.StringVar  # noqa: N815
        width: reflex.vars.NumberVar[float]
        height: reflex.vars.NumberVar[float]
        initialWidth: reflex.vars.NumberVar[float]  # noqa: N815
        initialHeight: reflex.vars.NumberVar[float]  # noqa: N815
        parentId: reflex.vars.StringVar  # noqa: N815
        zIndex: reflex.vars.NumberVar[int]  # noqa: N815
        extent: (
            reflex.vars.ArrayVar[CoordinateExtent]
            | reflex.vars.StringVar[Literal["parent"]]
        )
        expandParent: reflex.vars.BooleanVar  # noqa: N815
        ariaLabel: reflex.vars.StringVar  # noqa: N815
        origin: reflex.vars.ArrayVar[NodeOrigin]
        handles: reflex.vars.ArrayVar[Sequence[NodeHandleVar]]
        measured: MeasuredVar
        type: reflex.vars.StringVar[str | NodeType]
        style: reflex.vars.ObjectVar[Mapping[str, Any]]
        className: reflex.vars.StringVar  # noqa: N815
        resizing: reflex.vars.BooleanVar
        focusable: reflex.vars.BooleanVar
        ariaRole: reflex.vars.StringVar[AriaRole]  # noqa: N815
        domAttributes: reflex.vars.ObjectVar[Mapping[str, Any]]  # noqa: N815


class NodeHandleBoundsVar(reflex.vars.ObjectVar[NodeHandleBounds]):
    """Represents the bounds of a node's handles."""

    if TYPE_CHECKING:
        source: reflex.vars.ArrayVar[Sequence[NodeHandle]]
        target: reflex.vars.ArrayVar[Sequence[NodeHandle]]


class NodeBoundsVar(XYPositionVar):
    """Represents the bounds of a node."""

    if TYPE_CHECKING:
        width: reflex.vars.NumberVar[float]
        height: reflex.vars.NumberVar[float]


class NodeInternalsVar(reflex.vars.ObjectVar[NodeInternals]):
    """Represents the internal state of a node."""

    if TYPE_CHECKING:
        positionAbsolute: XYPositionVar  # noqa: N815
        z: reflex.vars.NumberVar[float]
        userNode: NodeVar  # noqa: N815
        handleBounds: NodeHandleBoundsVar | reflex.vars.base.NoneVar  # noqa: N815
        bounds: NodeBoundsVar | reflex.vars.base.NoneVar


class InternalNodeVar(NodeVar):
    """Represents an internal node in the flow."""

    if TYPE_CHECKING:
        internals: NodeInternalsVar


class EdgeMarkerVar(reflex.vars.ObjectVar[EdgeMarker]):
    """Represents a marker on an edge."""

    if TYPE_CHECKING:
        type: reflex.vars.StringVar[Literal["arrow", "arrowclosed"]]
        color: reflex.vars.StringVar | reflex.vars.base.NoneVar
        width: reflex.vars.NumberVar[float] | reflex.vars.base.NoneVar
        height: reflex.vars.NumberVar[float] | reflex.vars.base.NoneVar
        markerUnits: reflex.vars.StringVar | reflex.vars.base.NoneVar  # noqa: N815
        orient: reflex.vars.StringVar | reflex.vars.base.NoneVar
        strokeWidth: reflex.vars.NumberVar[float] | reflex.vars.base.NoneVar  # noqa: N815


class EdgeVar(reflex.vars.ObjectVar[Edge]):
    """Represents an edge in the flow."""

    if TYPE_CHECKING:
        id: reflex.vars.StringVar[str]
        type: reflex.vars.StringVar[EdgeType | str]
        source: reflex.vars.StringVar[str]
        target: reflex.vars.StringVar[str]
        sourceHandle: reflex.vars.StringVar[str] | reflex.vars.base.NoneVar  # noqa: N815
        targetHandle: reflex.vars.StringVar[str] | reflex.vars.base.NoneVar  # noqa: N815
        animated: reflex.vars.BooleanVar
        hidden: reflex.vars.BooleanVar
        deletable: reflex.vars.BooleanVar
        selectable: reflex.vars.BooleanVar
        data: reflex.vars.ObjectVar[Mapping[str, Any]]
        selected: reflex.vars.BooleanVar
        markerStart: reflex.vars.StringVar | EdgeMarkerVar  # noqa: N815
        markerEnd: reflex.vars.StringVar | EdgeMarkerVar  # noqa: N815
        zIndex: reflex.vars.NumberVar[float]  # noqa: N815
        ariaLabel: reflex.vars.StringVar[str]  # noqa: N815
        interactionWidth: reflex.vars.NumberVar[float]  # noqa: N815
        label: reflex.Var
        labelStyle: reflex.Var  # noqa: N815
        labelShowBg: reflex.vars.BooleanVar  # noqa: N815
        labelBgStyle: reflex.Var  # noqa: N815
        labelBgPadding: reflex.vars.ArrayVar[tuple[float, float]]  # noqa: N815
        labelBgBorderRadius: reflex.vars.NumberVar[float]  # noqa: N815
        style: reflex.Var
        className: reflex.vars.StringVar[str]  # noqa: N815
        reconnectable: reflex.vars.StringVar[HandleType] | reflex.vars.BooleanVar
        focusable: reflex.vars.BooleanVar
        ariaRole: reflex.vars.StringVar[AriaRole]  # noqa: N815
        domAttributes: reflex.vars.ObjectVar[Mapping[str, Any]]  # noqa: N815


class ConnectionVar(reflex.vars.ObjectVar[Connection]):
    """Represents a connection between two nodes in the flow."""

    if TYPE_CHECKING:
        source: reflex.vars.StringVar[str]
        target: reflex.vars.StringVar[str]
        sourceHandle: reflex.vars.StringVar[str] | reflex.vars.base.NoneVar  # noqa: N815
        targetHandle: reflex.vars.StringVar[str] | reflex.vars.base.NoneVar  # noqa: N815


class NodeConnectionVar(ConnectionVar):
    """Represents a connection between two nodes in the flow."""

    if TYPE_CHECKING:
        edgeId: reflex.vars.StringVar[str]  # noqa: N815


class ConnectionStateVar(reflex.vars.ObjectVar[ConnectionState]):
    """Represents the state of a connection between two nodes in the flow."""

    if TYPE_CHECKING:
        inProgress: reflex.vars.BooleanVar  # noqa: N815
        isValid: reflex.vars.BooleanVar | reflex.vars.base.NoneVar  # noqa: N815
        # from: XYPositionVar | reflex.vars.base.NoneVar
        fromHandle: NodeHandleVar | reflex.vars.base.NoneVar  # noqa: N815
        fromPosition: reflex.vars.StringVar[Position] | reflex.vars.base.NoneVar  # noqa: N815
        fromNode: NodeVar | reflex.vars.base.NoneVar  # noqa: N815
        to: XYPositionVar | reflex.vars.base.NoneVar
        toHandle: NodeHandleVar | reflex.vars.base.NoneVar  # noqa: N815
        toPosition: reflex.vars.StringVar[Position] | reflex.vars.base.NoneVar  # noqa: N815
        toNode: NodeVar | reflex.vars.base.NoneVar  # noqa: N815



"""Reflex state for OIDC authentication."""

import base64
import datetime
import hashlib
import logging
import secrets
import uuid
from typing import ClassVar
from urllib.parse import urlencode, urlparse

import httpx
import reflex as rx
from httpx._client import USE_CLIENT_DEFAULT
from joserfc.jwt import Token

from reflex_enterprise.app import AppEnterprise
from reflex_enterprise.auth.oidc.config import ConfigMixin
from reflex_enterprise.auth.oidc.types import OIDCUserInfo
from reflex_enterprise.auth.oidc.utils import compute_at_hash, verify_jwt
from reflex_enterprise.components.message_listener import (
    POST_MESSAGE_AND_CLOSE_POPUP,
    WINDOW_OPEN,
    WindowMessage,
    message_listener,
)
from reflex_enterprise.utils import call_event_from_computed_var

logger = logging.getLogger("reflex_enterprise.auth.oidc")
COOKIE_MAX_AGE = 604800


class OIDCAuthState(ConfigMixin, rx.State, mixin=True):
    """Reflex base state class for managing OIDC authentication flows.

    This state class handles the OAuth 2.0 Authorization Code flow with PKCE
    for OIDC authentication, including token storage, validation, and user
    information retrieval.

    Users should subclass this state and set the `__provider__` attribute
    to specify the OIDC provider (e.g., "okta", "databricks"), and supply
    appropriate environment variables for client ID, secret, and issuer URI.

    For example:
    ```python
    class OktaAuthState(OIDCAuthState, rx.State):
        __provider__ = "okta"
    ```

    Then use the `OktaAuthState.get_login_button()` method to render a login
    button in your app.

    Attributes:
        access_token: The OAuth 2.0 access token stored in cookie.
        id_token: The OpenID Connect ID token stored in cookie.
        granted_scopes: The scopes granted for the access_token stored in cookie.
        is_iframed: Whether the app is running inside an iframe.
        from_popup: Whether the current page was opened as a popup.

        _redirect_to_url: URL to redirect to after successful authentication.
        _app_state: Random state parameter for CSRF protection.
        _code_verifier: PKCE code verifier for secure authorization.
        _requested_scopes: Scopes requested during authentication.
        _last_error_message: Error message for authentication failures.
    """

    __provider__: ClassVar[str] = "generic"
    _has_registered_endpoints: ClassVar[bool] = False

    access_token: str = rx.Cookie(
        max_age=COOKIE_MAX_AGE,
        secure=True,
        same_site="strict",
    )
    id_token: str = rx.Cookie(
        max_age=COOKIE_MAX_AGE,
        secure=True,
        same_site="strict",
    )
    granted_scopes: str = rx.Cookie(
        max_age=COOKIE_MAX_AGE,
        secure=True,
        same_site="strict",
    )
    is_iframed: bool = False
    from_popup: bool = False

    _redirect_to_url: str
    _app_state: str
    _code_verifier: str
    _requested_scopes: str = "openid email profile"
    _nonce: str | None = None
    _expected_at_hash: str | None = None
    _last_error_message: str
    _last_error_txid: str

    @rx.event
    async def reset_auth(self):
        """Reset authentication state and clear tokens."""
        self.access_token = ""
        self.id_token = ""
        self.granted_scopes = ""

    async def _verify_jwt(self, token_json: str) -> Token:
        return await verify_jwt(
            token_json,
            issuer=await self._issuer_uri(),
            audience=await self._client_id(),
            nonce=self._nonce,
            at_hash=self._expected_at_hash,
        )

    async def _validate_tokens(self) -> bool:
        self._clear_last_error()

        if not self.access_token or not self.id_token:
            return False

        try:
            await self._verify_jwt(self.id_token)
        except Exception as e:
            self._set_last_error_message(f"ID token verification failed: {e}")
            return False

        return True

    def _clear_last_error(self):
        self._last_error_message = ""
        self._last_error_txid = ""

    def _set_last_error_message(self, msg: str):
        self._last_error_txid = uuid.uuid4().hex
        self._last_error_message = msg
        logger.info(
            f"{self.router.session.client_token} [txid={self._last_error_txid}] {self._last_error_message}"
        )

    @rx.var
    def has_error(self) -> bool:
        """Whether there was an authentication error."""
        return bool(self._last_error_message)

    @rx.var
    def last_error_txid(self) -> str:
        """Get the last error transaction ID for logging correlation."""
        return self._last_error_txid

    @rx.var(interval=datetime.timedelta(minutes=30))
    async def userinfo(self) -> OIDCUserInfo | None:
        """Get the authenticated user's information from OIDC token.

        This property retrieves the user's profile information from the OIDC
        userinfo endpoint using the stored access token. The result is cached
        for 30 minutes and automatically revalidated.

        Returns:
            OIDCUserInfo or subclass containing user profile data if authentication is valid,
            None if tokens are invalid or the request fails.
        """
        if not self.id_token:
            return None
        if not await self._validate_tokens():
            await call_event_from_computed_var(self, type(self).reset_auth)
            return None

        # Get the latest userinfo
        try:
            if userinfo_endpoint := await self._issuer_endpoint("userinfo_endpoint"):
                async with httpx.AsyncClient() as client:
                    resp = await client.get(
                        userinfo_endpoint,
                        headers={"Authorization": f"Bearer {self.access_token}"},
                    )
                    resp.raise_for_status()
                    return resp.json()  # pyright: ignore[reportReturnType]
        except Exception as e:
            self._set_last_error_message(f"Failed to fetch userinfo: {e}")
            return None
        # Have to just trust the ID token claims.
        return (await self._verify_jwt(self.id_token)).claims  # pyright: ignore[reportReturnType]

    def _redirect_uri(self) -> str:
        current_url = urlparse(self.router.url)
        return current_url._replace(
            path=self._authorization_code_endpoint(),
            query=None,
            fragment=None,
        ).geturl()

    def _index_uri(self) -> str:
        current_url = urlparse(self.router.url)
        return current_url._replace(path="/", query=None, fragment=None).geturl()

    @rx.event
    async def redirect_to_login_popup(self):
        """Open a small popup window to initiate the login flow.

        This is used when the app detects it's embedded and needs to open a
        dedicated popup for the authorization flow.
        """
        return rx.call_script(
            WINDOW_OPEN(
                self._popup_login_endpoint(),
                f"{self.__provider__}_popup_login",
                "width=600,height=600",
            )
        )

    @rx.event
    async def redirect_to_logout_popup(self):
        """Open a small popup window to initiate the logout flow."""
        self.access_token = self.id_token = self.granted_scopes = ""
        return rx.call_script(
            WINDOW_OPEN(
                self._popup_logout_endpoint(),
                f"{self.__provider__}_popup_login",
                "width=600,height=600",
            )
        )

    @rx.event
    def set_from_popup(self, from_popup: bool):
        """Set whether the current page was opened as a popup."""
        self.from_popup = from_popup

    @rx.event
    async def redirect_to_login(self):
        """Initiate the OAuth 2.0 authorization code flow with PKCE.

        This method generates the necessary state and code verifier for PKCE,
        constructs the authorization URL, and redirects the user to Okta's
        authorization endpoint.

        Returns:
            A redirect response to the Okta authorization endpoint.
        """
        if self.is_iframed:
            return type(self).redirect_to_login_popup()
        if await self._validate_tokens():
            return [
                self.post_auth_message(),
                rx.toast("You are logged in."),
            ]

        # store app state and code verifier in session
        self._app_state = secrets.token_urlsafe(64)
        self._code_verifier = secrets.token_urlsafe(64)
        self._redirect_to_url = self.router.url

        # calculate code challenge
        hashed = hashlib.sha256(self._code_verifier.encode("ascii")).digest()
        encoded = base64.urlsafe_b64encode(hashed)
        code_challenge = encoded.decode("ascii").strip("=")

        # get request params
        query_params = {
            "client_id": await self._client_id(),
            "redirect_uri": self._redirect_uri(),
            "scope": self._requested_scopes,
            "state": self._app_state,
            "code_challenge": code_challenge,
            "code_challenge_method": "S256",
            "response_type": "code",
            "response_mode": "query",
        }

        # build request_uri
        self._clear_last_error()
        try:
            request_uri = f"{await self._issuer_endpoint('authorization_endpoint')}?{urlencode(query_params)}"
        except Exception as e:
            self._set_last_error_message(f"Failed to build authorization URL: {e}")
            return rx.toast.error("Authentication error")
        return rx.redirect(request_uri)

    @rx.event
    async def redirect_to_logout(self):
        """Initiate the OAuth 2.0 logout flow.

        This method generates a new state parameter, constructs the logout URL
        with the ID token hint, and redirects the user to Okta's logout endpoint.
        The user's tokens are cleared from local storage after the redirect.

        Returns:
            A redirect response to the Okta logout endpoint.
        """
        if self.is_iframed:
            return type(self).redirect_to_logout_popup()

        # store app state and code verifier in session
        self._app_state = secrets.token_urlsafe(64)

        # get request params
        query_params = {
            "id_token_hint": self.id_token,
            "state": self._app_state,
        }
        if not self.from_popup:
            query_params["post_logout_redirect_uri"] = self._index_uri()

        self._clear_last_error()
        try:
            if end_session_endpoint := await self._issuer_endpoint(
                "end_session_endpoint"
            ):
                request_uri = f"{end_session_endpoint}?{urlencode(query_params)}"
                return rx.redirect(request_uri)
        except Exception as e:
            self._set_last_error_message(f"Failed to build logout URL: {e}")
            return rx.toast.error("Logout error")
        finally:
            await self.reset_auth()

    @rx.event
    async def auth_callback(self):
        """Handle the OAuth 2.0 authorization-code callback.

        This method is called when the user is redirected back from OIDC
        authorization endpoint. It validates the state parameter to prevent CSRF
        attacks, exchanges the authorization code for tokens using PKCE, and
        stores the tokens for future use.

        Returns:
            A redirect response to the original requested URL, or an error toast
            if authentication fails.
        """
        self._clear_last_error()
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        code = self.router.url.query_parameters.get("code")
        app_state = self.router.url.query_parameters.get("state")
        if app_state != self._app_state:
            self._set_last_error_message("App state mismatch. Possible CSRF attack.")
            return rx.toast.error("Authentication error")
        if not code:
            self._set_last_error_message("No code provided in the callback.")
            return rx.toast.error("Authentication error")
        query_params = {
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": self._redirect_uri(),
            "code_verifier": self._code_verifier,
            "scope": self._requested_scopes,
        }
        if client_secret := await self._client_secret():
            auth = (await self._client_id(), client_secret)
        else:
            query_params["client_id"] = await self._client_id()
            auth = USE_CLIENT_DEFAULT
        try:
            async with httpx.AsyncClient() as client:
                resp = await client.post(
                    await self._issuer_endpoint("token_endpoint"),
                    headers=headers,
                    data=query_params,
                    auth=auth,
                )
                resp.raise_for_status()
                exchange = resp.json()
        except Exception as e:
            self._set_last_error_message(f"Token exchange failed: {e}")
            return rx.toast.error("Authentication error")

        # Get tokens and validate
        if exchange.get("token_type") != "Bearer":
            self._set_last_error_message("Unsupported token type. Should be 'Bearer'.")
            return rx.toast.error("Authentication error")
        await self._set_tokens(
            access_token=exchange["access_token"],
            id_token=exchange["id_token"],
            granted_scopes=exchange.get("scope", ""),
        )

        return rx.redirect(self._redirect_to_url)

    async def _set_tokens(
        self, access_token: str, id_token: str, granted_scopes: str = ""
    ):
        self._nonce = None
        self._expected_at_hash = None
        self.access_token = access_token
        self.id_token = id_token
        self.granted_scopes = granted_scopes

        # compute at_hash and store for additional validation
        try:
            token = await self._verify_jwt(id_token)
            alg = token.header.get("alg")
            at_hash = compute_at_hash(self.access_token, alg)
            # store expected at_hash in a transient property (not persisted)
            self._expected_at_hash = at_hash
        except Exception:
            self._expected_at_hash = None

        # Validate again after setting expected_at_hash
        if not await self._validate_tokens():
            self.access_token = self.id_token = self.granted_scopes = ""

    @rx.var
    def origin(self) -> str:
        """Return the app origin URL (used as postMessage target origin)."""
        return self._index_uri().rstrip("/")

    @rx.event
    def check_if_iframed(self):
        """Run a short client-side check to determine whether the page is iframed.

        The result is reported to `check_if_iframed_cb`.
        """
        return rx.call_function(
            """() => {
    try {
        return window.self !== window.top;
    } catch (e) {
        // This catch block handles potential security errors (Same-Origin Policy)
        // if the iframe content and the parent are from different origins.
        // In such cases, access to window.top might be restricted, implying it's in an iframe.
        return true;
    }
}""",
            callback=type(self).check_if_iframed_cb,
        )

    @rx.event
    def check_if_iframed_cb(self, is_iframed: bool):
        """Callback invoked with the iframe detection result.

        Args:
            is_iframed: True if the page is inside an iframe or cross-origin
                access prevented detection.
        """
        self.is_iframed = is_iframed

    @rx.event
    async def on_iframe_auth_success(self, event: WindowMessage):
        """Handle an authentication success message posted from a child window.

        The message payload is expected to include `access_token`, `id_token`,
        and an optional `nonce`. Tokens are stored via `_set_tokens`.
        """
        if event["data"].get("type") != "auth":
            return
        await self._set_tokens(
            access_token=event["data"].get("access_token"),
            id_token=event["data"].get("id_token"),
            granted_scopes=event["data"].get("scope", ""),
        )

    @rx.event
    def post_auth_message(self):
        """Post tokens back to the opening window and close the popup.

        This is called on the popup page when authentication has completed and
        the tokens are available in `self.access_token` / `self.id_token`.
        """
        payload = {
            "type": "auth",
            "access_token": self.access_token,
            "id_token": self.id_token,
            "scope": self.granted_scopes,
        }
        return rx.call_script(POST_MESSAGE_AND_CLOSE_POPUP(payload, self.origin, 500))

    @classmethod
    def get_login_button(cls, *children) -> rx.Component:
        """Return a login button component that initiates OIDC auth.

        If `children` are provided they will be placed inside the clickable
        element; otherwise a default button label is used. The component wires up
        the message listener (for iframe flows), the click handler, and a mount
        handler that checks whether the page is embedded in an iframe.
        """
        cls.register_auth_endpoints()
        if not children:
            children = [rx.button(f"Login with {cls.__provider__.title()}")]
        return rx.el.div(
            *children,
            rx.cond(
                cls.is_iframed,
                message_listener(
                    allowed_origin=cls.origin,
                    on_message=cls.on_iframe_auth_success,
                ),
            ),
            on_click=cls.redirect_to_login,
            on_mount=cls.check_if_iframed,
            width="fit-content",
        )

    @classmethod
    def get_state_hydrating_component(cls) -> rx.Component:
        """Loading spinner shown before state is hydrated."""
        return rx.vstack(
            rx.spinner(),
            align="center",
            justify="center",
            height="50vh",
            width="100%",
        )

    @classmethod
    def _with_hydrated(cls, *components: rx.Component) -> rx.Component:
        """Wrap components to wait for state hydration before rendering.

        Args:
            components: The components to render after hydration.

        Returns:
            A component that shows a loading spinner until state is hydrated,
            then renders the provided components.
        """
        return rx.cond(
            rx.State.is_hydrated,
            rx.fragment(*components),
            cls.get_state_hydrating_component(),
        )

    @classmethod
    def get_authentication_loading_page(cls) -> rx.Component:
        """Small loading page shown while authentication is validated.

        This page is registered by the package as the callback target when the
        authorization response is being processed.
        """
        return rx.container(
            cls._with_hydrated(
                rx.cond(
                    cls.has_error,
                    rx.vstack(
                        rx.heading("An error occurred during authentication."),
                        rx.text(
                            "Please close this window and try again.",
                        ),
                        rx.text(
                            "An administrator may provide more information about error ID: ",
                            rx.badge(cls.last_error_txid),
                        ),
                    ),
                    rx.cond(
                        ~cls.userinfo,
                        rx.hstack(
                            rx.heading("Validating Authentication..."),
                            rx.spinner(),
                            width="50%",
                            justify="between",
                        ),
                        rx.heading("Redirecting to app..."),
                    ),
                ),
            ),
        )

    @classmethod
    def get_authentication_popup_logout(cls) -> rx.Component:
        """Simple page shown during the logout flow.

        Registered at `/_reflex_oidc_{provider}/popup-logout` to complete the sign-out handshake.
        """
        return rx.container(
            cls._with_hydrated(
                rx.cond(
                    cls.has_error,
                    rx.vstack(
                        rx.heading("An error occurred during logout."),
                        rx.text(
                            "You close this window and clear browser cookies to manually log out.",
                        ),
                        rx.text(
                            "An administrator may provide more information about error ID: ",
                            rx.badge(cls.last_error_txid),
                        ),
                    ),
                    rx.cond(
                        cls.id_token,
                        rx.hstack(
                            rx.heading("Complete logout process."),
                            rx.spinner(),
                            width="50%",
                            justify="between",
                        ),
                        rx.heading("You are logged out. You may close this window."),
                    ),
                ),
            ),
        )

    @classmethod
    def register_auth_endpoints(cls, app: AppEnterprise | None = None):
        """Register the Okta authentication endpoints with the Reflex app.

        This function sets up the necessary OAuth callback endpoint for handling
        authentication responses from OIDC providers. The callback endpoint
        handles the authorization code exchange and redirects users.
        """
        if app is None and cls._has_registered_endpoints:
            return
        if app is None:
            cls._has_registered_endpoints = True

            from reflex.utils.prerequisites import get_app

            app = get_app().app
        if not isinstance(app, AppEnterprise):
            raise TypeError("The app must be an instance of reflex_enterprise.App.")
        app.add_page(
            cls.get_authentication_loading_page(),
            route=cls._authorization_code_endpoint(),
            on_load=cls.auth_callback,
            title=f"{cls.__provider__.title()} Auth Callback",
        )
        app.add_page(
            cls.get_authentication_loading_page(),
            route=cls._popup_login_endpoint(),
            on_load=[
                cls.set_from_popup(True),
                cls.redirect_to_login,
            ],
            title=f"{cls.__provider__.title()} Auth Initiator",
        )
        app.add_page(
            cls.get_authentication_popup_logout(),
            route=cls._popup_logout_endpoint(),
            on_load=[
                cls.set_from_popup(True),
                cls.redirect_to_logout,
            ],
            title=f"{cls.__provider__.title()} Auth Logout",
        )



"""Utility functions for Reflex Enterprise."""

import dataclasses
import inspect
from typing import Any, Callable

import psutil
from reflex import ImportVar, constants
from reflex.config import get_config
from reflex.event import EventCallback, EventHandler, EventSpec, fix_events
from reflex.state import BaseState
from reflex.utils import console, prerequisites
from reflex.utils.decorator import once
from reflex.utils.prerequisites import get_app
from reflex.vars.base import Var, VarData
from reflex.vars.function import ArgsFunctionOperation, FunctionStringVar

from reflex_enterprise.constants import IS_OFFLINE

from .vars import ArgsFunctionOperationPromise, PromiseVar


def get_user_tier() -> str:
    """Get the current user's subscription tier.

    Returns:
        The user's subscription tier as a string.
    """
    if IS_OFFLINE:
        return "enterprise"
    return prerequisites.get_user_tier() or "anonymous"


def check_config_option_in_tier(
    option_name: str,
    allowed_tiers: list[str],
    fallback_value: Any,
    help_link: str | None = None,
):
    """Check if a config option is allowed for the authenticated user's current tier.

    Args:
        option_name: The name of the option to check.
        allowed_tiers: The tiers that are allowed to use the option.
        fallback_value: The fallback value if the option is not allowed.
        help_link: The help link to show to a user that is authenticated.
    """
    config = get_config()
    current_tier = get_user_tier()

    if allowed_tiers == []:
        the_remedy = "This option is not available in the current context."
    else:
        if current_tier == "anonymous":
            the_remedy = "You are currently logged out. Run `reflex login` to access this option."
        else:
            the_remedy = (
                f"Your current subscription tier is `{current_tier}`. "
                f"Please upgrade to {allowed_tiers} to access this option. "
            )
            if help_link:
                the_remedy += f"See {help_link} for more information."

    value = getattr(config, option_name)

    if value is None:
        setattr(config, option_name, fallback_value)
        return

    if value != fallback_value and current_tier not in allowed_tiers:
        console.warn(f"Config option `{option_name}` is restricted. {the_remedy}")
        setattr(config, option_name, fallback_value)
        config._set_persistent(**{option_name: fallback_value})


def is_deploy_context():
    """Check if the current context is a deploy context.

    Returns:
        True if the current context is a deploy context, False otherwise.
    """
    from reflex.utils.exec import get_compile_context

    return get_compile_context() == constants.CompileContext.DEPLOY  # pyright: ignore [reportPrivateImportUsage]


def get_backend_url(relative_url: str | Var[str]) -> Var[str]:
    """Get the full backend URL for a given relative URL.

    Use with `fetch` to access backend API endpoints.

    Args:
        relative_url: The relative URL to convert.

    Returns:
        A Var representing the full backend URL.
    """
    return ArgsFunctionOperation.create(
        args_names=("url",),
        return_expr=Var(
            r"`${getBackendURL(typeof env !== 'undefined' ? env.UPLOAD : env_default.UPLOAD).origin}/${url.replace(/^\/+/, '')}`"
        ),
        _var_data=VarData(
            imports={
                "$/utils/state": ["getBackendURL"],
                "$/env.json": ImportVar(tag="env", is_default=True),
            }
        ),
    )(relative_url).to(str)


def fetch(
    url: str | Var[str], options: dict[str, Any] | Var[dict[str, Any]] | None = None
) -> PromiseVar:
    """Fetch a URL with the given options.

    Args:
        url: The URL to fetch.
        options: The options to use for the fetch.

    Returns:
        A PromiseVar representing the eventual result of the fetch.
    """
    return ArgsFunctionOperationPromise.create(
        args_names=("url", "options"),
        return_expr=Var(
            "fetch(url, {...options, headers: {...options.headers, 'X-Reflex-Client-Token': getToken()}})"
        ),
        _var_data=VarData(
            imports={
                "$/utils/state": [
                    "getToken",
                ],
            }
        ),
    )(url, options or {})


def encode_uri_component(value: str | Var[str]) -> Var[str]:
    """Encode a URI component.

    Args:
        value: The value to encode.

    Returns:
        A Var representing the encoded URI component.
    """
    return FunctionStringVar.create("encodeURIComponent")(value).to(str)


def arrow_func(py_fn: Callable) -> ArgsFunctionOperation:
    """Convert a Python function to a js arrow function."""
    sig = inspect.signature(py_fn)
    params = [
        Var(
            param.name,
            _var_type=param.annotation,
        ).guess_type()
        for param in sig.parameters.values()
    ]
    return ArgsFunctionOperation.create(
        args_names=tuple(sig.parameters),
        return_expr=py_fn(*params),
    )


@once
def is_new_session() -> bool:
    """Check if this is a new development session vs hot reload restart.

    Returns:
        True if this is a new session, False if it's a hot reload restart.
    """
    # Get current parent PID (this could fail in unusual environments)
    try:
        current_parent_pid = (
            parent.pid if (parent := psutil.Process().parent()) else None
        )
    except Exception:
        return True  # Default to showing message if process detection fails

    if current_parent_pid is None:
        return True

    # Check stored parent PID from previous session
    backend_dir = prerequisites.get_web_dir() / "backend"
    backend_dir.mkdir(parents=True, exist_ok=True)
    session_file = backend_dir / "session_parent_pid"

    if session_file.exists():
        # Read and parse stored PID (this could fail if file is corrupted)
        try:
            stored_parent_pid = int(session_file.read_text().strip())
            is_new = current_parent_pid != stored_parent_pid
        except Exception:
            # File is corrupted, treat as new session
            is_new = True
    else:
        # First time, treat as new session
        is_new = True

    if is_new:
        session_file.write_text(str(current_parent_pid))

    return is_new


async def call_event_from_computed_var(
    state: BaseState,
    event: EventSpec | EventHandler | EventCallback,
):
    """Chain an event from an async computed var.

    Args:
        state: The state to call the event on.
        event: The event to call.
    """
    token = state.router.session.client_token
    fixed_event = fix_events(
        events=[event],  # pyright: ignore[reportArgumentType]
        token=token,
        router_data=state.router_data,
    )[0]
    substate, handler = state._get_root_state()._get_event_handler(fixed_event)

    if handler.is_background:
        msg = "Background events cannot be chained from computed vars."
        raise RuntimeError(msg)

    app = get_app().app

    async for update in state._process_event(
        handler=handler,
        state=substate,
        payload=fixed_event.payload,
    ):
        # In-bound updates are not final! The delta that triggered the computed
        # var may be final and we don't want to override that.
        update = dataclasses.replace(
            update,
            final=False,
        )
        # Send the update to the client.
        await app.event_namespace.emit_update(
            update=update,
            token=token,
        )
